import re
from collections import namedtuple

from robot.api import get_model, Token
from robot.parsing import ModelVisitor


class ParserApplication(ModelVisitor):
    """ Class used to extract all useful info from the .robot file.

    See https://robot-framework.readthedocs.io/en/v6.0.2/autodoc/robot.api.html#inspecting-model

    Attributes:
        tests (list): List of objects with attributes name (str), doc (str) and tags (list).
        variables (dict): Dictionary mapping variables, e.g. '${MESSAGE}', to their values
    """
    TestAttributes = namedtuple('TestAttributes', 'name doc tags')

    def __init__(self, robot_file, tags_for_inclusion, *args, **kwargs):
        """ Constructor

        Args:
            robot_file (Path): Path to the .robot file.
            tags_for_inclusion (list): Regexes for matching tags. A test case is included if every regex matches at
                least one of its tags.
        """
        super().__init__(*args, **kwargs)
        self.robot_file = str(robot_file.resolve(strict=True))
        self.model = get_model(self.robot_file)
        self.tests = []
        self.variables = {}
        self.tags_for_inclusion = tags_for_inclusion

    def run(self):
        self.visit(self.model)

    def visit_VariableSection(self, node):
        for element in node.body:
            element_type = getattr(element, 'type', None)
            if element_type == Token.VARIABLE:
                name = element.get_value(Token.VARIABLE)
                value = ' '.join(element.get_values(Token.ARGUMENT))
                match = re.fullmatch(r"\${(.+)}", value)
                if match:
                    value = match.group(1)
                self.variables[name] = value

    def visit_TestCase(self, node):
        doc = ''
        tags = []
        for element in node.body:
            element_type = getattr(element, 'type', None)
            if element_type == Token.DOCUMENTATION:
                in_docstring = False
                previous_token = None
                for token in element.tokens:
                    if in_docstring and token.type in (Token.ARGUMENT, Token.EOL, Token.SEPARATOR):
                        if previous_token is None or previous_token.type != Token.CONTINUATION:
                            doc += token.value
                        elif len(token.value) > 2:
                            doc += token.value[2:]  # remove two leading spaces, which are needed to separate the text
                    elif token.type == Token.CONTINUATION:
                        doc = doc.rstrip(' ')
                    elif token.type == Token.DOCUMENTATION:
                        in_docstring = True
                    previous_token = token
            elif element_type == Token.TAGS:
                tags = [el.value for el in element.tokens if el.type == Token.ARGUMENT]
        if self.evaluate_inclusion(tags):
            self.tests.append(self.TestAttributes(node.name, doc, tags))

    def evaluate_inclusion(self, tags):
        for pattern in self.tags_for_inclusion:
            regexp = re.compile(pattern)
            for tag in tags:
                if regexp.match(tag):
                    break
            else:
                return False
        return True



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

project_url = 'https://github.com/melexis/robot2rst'

requires = ['robotframework>=3.2', 'mako']

setup(
    name='mlx.robot2rst',
    use_scm_version={'write_to': 'mlx/__robot2rst_version__.py'},
    url=project_url,
    author='Jasper Craeghs',
    author_email='jce@melexis.com',
    license='Apache License Version 2.0',
    description='Python script for converting a Robot Framework file to a reStructuredText (.rst) file',
    long_description=open("README.rst").read(),
    long_description_content_type='text/x-rst',
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Documentation',
        'Topic :: Documentation :: Sphinx',
        'Topic :: Utilities',
    ],
    platforms='any',
    packages=find_packages(exclude=['tests', 'doc']),
    include_package_data=True,
    install_requires=requires,
    setup_requires=['setuptools_scm'],
    python_requires='>=3.7',
    namespace_packages=['mlx'],
    keywords=['robot', 'robotframework', 'sphinx', 'traceability'],
    entry_points={
        'console_scripts': [
            'mlx.robot2rst = mlx.robot2rst:entrypoint',
            'robot2rst = mlx.robot2rst:entrypoint',
        ]
    },
)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import mlx.traceability
from pkg_resources import get_distribution

# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('../mlx'))

# -- Project information -----------------------------------------------------

project = 'mlx.robot2rst'
copyright = '2020, Jasper Craeghs'
authors = ['Stein Heselmans', 'Jasper Craeghs']

# The full version, including alpha/beta/rc tags
release = get_distribution('mlx.robot2rst').version
version = '.'.join(release.split('.')[:2])

man_pages = [
    ('index', 'robot2rst', 'Script to convert .robot files to .rst files with traceable items',
     authors, 1)
]

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'mlx.traceability',
    'sphinx_selective_exclude.eager_only',
]

traceability_relationships = {
    'validates': 'validated_by',
    'ext_toolname': '',
}
traceability_relationship_to_string = {
    'validates': 'Validates',
    'validated_by': 'Validated by',
    'ext_toolname': 'Reference to toolname'
}
traceability_external_relationship_to_url = {
    'ext_toolname': 'http://toolname.company.com/my_lib/system-requirements.html#field1'
}
traceability_render_relationship_per_item = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = [os.path.join(os.path.dirname(mlx.traceability.__file__), 'assets')]


def setup(app):
    tags.add('FLASH')




''' Script to convert a robot test file to a reStructuredText file with traceable items '''
import argparse
import logging
import sys
from textwrap import indent
from pathlib import Path

from mako.exceptions import RichTraceback
from mako.template import Template

from .robot_parser import ParserApplication

TEMPLATE_FILE = Path(__file__).parent.joinpath('robot2rst.mako')
LOGGER = logging.getLogger(__name__)


def render_template(destination, only="", **kwargs):
    """
    Renders the Mako template, and writes output file to the specified destination.

    Args:
        destination (Path): Location of the output file.
        only (str): Expression for 'only' directive, which will only be added when this string is not empty.
        **kwargs (dict): Variables to be used in the Mako template.
    Raises:
        CRITICAL: Error was raised by Mako template.
    """
    destination.parent.mkdir(parents=True, exist_ok=True)
    template = Template(filename=str(TEMPLATE_FILE))
    try:
        rst_content = template.render(**kwargs)
    except Exception as err:
        traceback = RichTraceback()
        LOGGER.critical("File %s, line %s, in %s: %r", *traceback.traceback[-1])
        LOGGER.critical(repr(err))
        return 1
    else:
        if only:
            rst_content = f".. only:: {only}\n\n{indent(rst_content, ' ' * 4)}"
        with open(str(destination), 'w', encoding='utf-8', newline='\n') as rst_file:
            rst_file.write(rst_content)
    return 0


def generate_robot_2_rst(parser, rst_file, prefix, relationship_config, gen_matrix, **kwargs):
    """
    Calls mako template function and passes all needed parameters.

    Args:
        parser (ParserApplication): Parser with data extracted from the .robot file
        rst_file (Path): Path to the output file (.rst).
        prefix (str): Prefix of generated item IDs.
        relationship_config (list): List of tuples that contain a relationship, tag_regex and coverage percentage
        gen_matrix (bool): True if traceability matrices are to be generated, False if not.
    """
    return render_template(
        rst_file,
        parser=parser,
        suite=rst_file.stem,
        prefix=prefix,
        relationship_config=relationship_config,
        gen_matrix=gen_matrix,
        **kwargs,
    )


def _tweak_prefix(prefix):
    """ If a prefix or regex ends with '_-', change it to end with '-' instead.

    Args:
        prefix (str): Prefix that might need tweaking.

    Returns:
        (str) Prefix that has been tweaked if it needed to.
    """
    if prefix.endswith('_-'):
        prefix = prefix.rstrip('_-') + '-'
    return prefix


def main():
    '''Main entry point for script: parse arguments and execute'''
    parser = argparse.ArgumentParser(description='Convert robot test cases to reStructuredText with traceable items.')
    parser.add_argument("-i", "--robot", dest='robot_file', required=True,
                        help='Input robot file')
    parser.add_argument("-o", "--rst", dest='rst_file', required=True,
                        help='Output RST file, e.g. my_component_qtp.rst')
    parser.add_argument("--only", dest="expression", default="",
                        help="Expression of tags for Sphinx' `only` directive that surrounds all RST content. "
                        "By default, no `only` directive is generated.")
    parser.add_argument("-p", "--prefix", default='QTEST-',
                        help="Overrides the default 'QTEST-' prefix.")
    parser.add_argument("-r", "--relationships", nargs='*',
                        help="Name(s) of the relationship(s) used to link to items in Tags section. The default value "
                             "is 'validates'.")
    parser.add_argument("-t", "--tags", nargs='*',
                        help="Zero or more Python regexes for matching tags to treat them as traceable targets via a "
                             "relationship. All tags get matched by default.")
    parser.add_argument("--include", nargs='*', default=[],
                        help="Zero or more Python regexes for matching tags to filter test cases. "
                             "If every regex matches at least one of a test case's tags, the test case is included.")
    parser.add_argument("-c", "--coverage", nargs='*',
                        help="Minimum coverage percentages for the item-matrix(es); 1 value per tag in -t, --tags.")
    parser.add_argument("--type", default='q',
                        help="Give value that starts with 'q' or 'i' (case-insensitive) to explicitly define "
                             "the type of test: qualification/integration test. The default is 'qualification'.")
    parser.add_argument("--trim-suffix", action='store_true',
                        help="If the suffix of any prefix or --tags argument ends with '_-' it gets trimmed to '-'.")

    logging.basicConfig(level=logging.INFO)
    args = parser.parse_args()

    type_map = {
        'i': 'integration',
        'q': 'qualification',
    }
    if args.type and args.type.lower()[0] in type_map:
        test_type = type_map[args.type.lower()[0]]
    else:
        raise ValueError(f"The --type argument is invalid: expected a value that starts with {' or '.join(type_map)}; "
                         f"got {args.type.lower()!r}.")

    relationships = args.relationships if args.relationships else ['validates']
    coverages = args.coverage if args.coverage else [0] * len(relationships)
    prefix = _tweak_prefix(args.prefix) if args.trim_suffix else args.prefix
    gen_matrix = True
    if not args.tags:
        args.tags = ['.*']  # only one relationship shall be used
        gen_matrix = False
        LOGGER.warning("No traceability matrix will be generated because of the use of default tag regex %r.",
                       args.tags[0])
    tag_regexes = [_tweak_prefix(regex) if args.trim_suffix else regex for regex in args.tags]

    if len(relationships) != len(tag_regexes):
        raise ValueError(f"Number of relationships ({len(relationships)}) is not equal to number of tag regexes "
                         f"({len(tag_regexes)}).")
    if len(relationships) != len(coverages):
        raise ValueError(f"Number of relationships ({len(relationships)}) is not equal to number of coverage "
                         f"percentages ({len(coverages)}).")
    relationship_config = [(relationships[i], tag_regexes[i], coverages[i]) for i in range(len(relationships))]

    parser = ParserApplication(Path(args.robot_file), args.include)
    parser.run()
    if parser.tests:
        exit_code = generate_robot_2_rst(parser, Path(args.rst_file), prefix, relationship_config,
                                         gen_matrix, test_type=test_type, only=args.expression, coverages=coverages)
    else:
        msg = f"robot2rst did not generate {args.rst_file} because {args.robot_file} does not contain any test cases"
        if args.include:
            msg += f" with tags matching all regexes {args.include}"
        LOGGER.info(msg)
        exit_code = 0
    return exit_code


def entrypoint():
    sys.exit(main())


if __name__ == "__main__":
    entrypoint()
